﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [StorableClass]
  [Item("SymbolicDataAnalysisExpressionTreeLinearInterpreter", "Interpreter for symbolic expression trees including automatically defined functions.")]
  public class SymbolicDataAnalysisExpressionTreeLinearInterpreter : ParameterizedNamedItem, ISymbolicDataAnalysisExpressionTreeInterpreter {
    private const string CheckExpressionsWithIntervalArithmeticParameterName = "CheckExpressionsWithIntervalArithmetic";
    private const string EvaluatedSolutionsParameterName = "EvaluatedSolutions";

    public override bool CanChangeName { get { return false; } }
    public override bool CanChangeDescription { get { return false; } }

    #region parameter properties
    public IValueParameter<BoolValue> CheckExpressionsWithIntervalArithmeticParameter {
      get { return (IValueParameter<BoolValue>)Parameters[CheckExpressionsWithIntervalArithmeticParameterName]; }
    }

    public IValueParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (IValueParameter<IntValue>)Parameters[EvaluatedSolutionsParameterName]; }
    }
    #endregion

    #region properties
    public BoolValue CheckExpressionsWithIntervalArithmetic {
      get { return CheckExpressionsWithIntervalArithmeticParameter.Value; }
      set { CheckExpressionsWithIntervalArithmeticParameter.Value = value; }
    }

    public IntValue EvaluatedSolutions {
      get { return EvaluatedSolutionsParameter.Value; }
      set { EvaluatedSolutionsParameter.Value = value; }
    }
    #endregion

    [StorableConstructor]
    protected SymbolicDataAnalysisExpressionTreeLinearInterpreter(bool deserializing) : base(deserializing) { }
    protected SymbolicDataAnalysisExpressionTreeLinearInterpreter(SymbolicDataAnalysisExpressionTreeLinearInterpreter original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicDataAnalysisExpressionTreeLinearInterpreter(this, cloner);
    }

    public SymbolicDataAnalysisExpressionTreeLinearInterpreter()
      : base("SymbolicDataAnalysisExpressionTreeLinearInterpreter", "Interpreter for symbolic expression trees including automatically defined functions.") {
      Parameters.Add(new ValueParameter<BoolValue>(CheckExpressionsWithIntervalArithmeticParameterName, "Switch that determines if the interpreter checks the validity of expressions with interval arithmetic before evaluating the expression.", new BoolValue(false)));
      Parameters.Add(new ValueParameter<IntValue>(EvaluatedSolutionsParameterName, "A counter for the total number of solutions the interpreter has evaluated", new IntValue(0)));
    }

    protected SymbolicDataAnalysisExpressionTreeLinearInterpreter(string name, string description)
      : base(name, description) {
      Parameters.Add(new ValueParameter<BoolValue>(CheckExpressionsWithIntervalArithmeticParameterName, "Switch that determines if the interpreter checks the validity of expressions with interval arithmetic before evaluating the expression.", new BoolValue(false)));
      Parameters.Add(new ValueParameter<IntValue>(EvaluatedSolutionsParameterName, "A counter for the total number of solutions the interpreter has evaluated", new IntValue(0)));
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(EvaluatedSolutionsParameterName))
        Parameters.Add(new ValueParameter<IntValue>(EvaluatedSolutionsParameterName, "A counter for the total number of solutions the interpreter has evaluated", new IntValue(0)));
    }

    #region IStatefulItem
    public void InitializeState() {
      EvaluatedSolutions.Value = 0;
    }

    public void ClearState() {
    }
    #endregion

    public IEnumerable<double> GetSymbolicExpressionTreeValues(ISymbolicExpressionTree tree, Dataset dataset, IEnumerable<int> rows) {
      if (CheckExpressionsWithIntervalArithmetic.Value)
        throw new NotSupportedException("Interval arithmetic is not yet supported in the symbolic data analysis interpreter.");

      lock (EvaluatedSolutions) {
        EvaluatedSolutions.Value++; // increment the evaluated solutions counter
      }

      Instruction[] code = SymbolicExpressionTreeCompiler.CompilePostfix(tree, OpCodes.MapSymbolToOpCode);
      foreach (Instruction instr in code) {
        switch (instr.opCode) {
          case OpCodes.Variable: {
              var variableTreeNode = (VariableTreeNode)instr.dynamicNode;
              instr.iArg0 = dataset.GetReadOnlyDoubleValues(variableTreeNode.VariableName);
            }
            break;
          case OpCodes.LagVariable: {
              var laggedVariableTreeNode = (LaggedVariableTreeNode)instr.dynamicNode;
              instr.iArg0 = dataset.GetReadOnlyDoubleValues(laggedVariableTreeNode.VariableName);
            }
            break;
          case OpCodes.VariableCondition: {
              var variableConditionTreeNode = (VariableConditionTreeNode)instr.dynamicNode;
              instr.iArg0 = dataset.GetReadOnlyDoubleValues(variableConditionTreeNode.VariableName);
            }
            break;
        }
      }

      double[] values = new double[code.Length];

      foreach (var rowEnum in rows) {
        int row = rowEnum;
        yield return Evaluate(dataset, ref row, code, values);
      }
    }

    protected virtual double Evaluate(Dataset dataset, ref int row, Instruction[] code, double[] values) {
      int count = 0;
      for (int j = 0; j != code.Length; ++j) {
        Instruction currentInstr = code[j];

        #region switch
        switch (currentInstr.opCode) {
          case OpCodes.Add: {
              double s = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                s += values[i];
              }
              values[j] = s;
            }
            break;
          case OpCodes.Sub: {
              double s = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                s -= values[i];
              }
              if (currentInstr.nArguments == 1) s = -s;
              values[j] = s;
            }
            break;
          case OpCodes.Mul: {
              double p = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                p *= values[i];
              }
              values[j] = p;
            }
            break;
          case OpCodes.Div: {
              double p = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                p /= values[i];
              }
              if (currentInstr.nArguments == 1) p = 1.0 / p;
              values[j] = p;
            }
            break;
          case OpCodes.Average: {
              double sum = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                sum += values[i];
              }
              values[j] = sum / currentInstr.nArguments;
            }
            break;
          case OpCodes.Cos: {
              values[j] = Math.Cos(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Sin: {
              values[j] = Math.Sin(values[j - currentInstr.nArguments]);
              break;
            }
          case OpCodes.Tan: {
              values[j] = Math.Tan(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Square: {
              values[j] = Math.Pow(values[j - currentInstr.nArguments], 2);
            }
            break;
          case OpCodes.Power: {
              double x = values[j - currentInstr.nArguments];
              double y = Math.Round(values[j - currentInstr.nArguments + 1]);
              values[j] = Math.Pow(x, y);
            }
            break;
          case OpCodes.SquareRoot: {
              values[j] = Math.Sqrt(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Root: {
              double x = values[j - currentInstr.nArguments];
              double y = Math.Round(values[j - currentInstr.nArguments + 1]);
              values[j] = Math.Pow(x, 1 / y);
            }
            break;
          case OpCodes.Exp: {
              values[j] = Math.Exp(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Log: {
              values[j] = Math.Log(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Gamma: {
              values[j] = double.IsNaN(values[j - currentInstr.nArguments]) ? double.NaN : alglib.gammafunction(values[j - currentInstr.nArguments]);
            }
            break;
          case OpCodes.Psi: {
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else if (x <= 0 && (Math.Floor(x) - x).IsAlmost(0)) values[j] = double.NaN;
              else values[j] = alglib.psi(x);
            }
            break;
          case OpCodes.Dawson: {
              var x = values[j - currentInstr.nArguments];
              values[j] = double.IsNaN(x) ? double.NaN : alglib.dawsonintegral(x);
            }
            break;
          case OpCodes.ExponentialIntegralEi: {
              var x = values[j - currentInstr.nArguments];
              values[j] = double.IsNaN(x) ? double.NaN : alglib.exponentialintegralei(x);
            }
            break;
          case OpCodes.SineIntegral: {
              double si, ci;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.sinecosineintegrals(x, out si, out ci);
                values[j] = si;
              }
            }
            break;
          case OpCodes.CosineIntegral: {
              double si, ci;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.sinecosineintegrals(x, out si, out ci);
                values[j] = ci;
              }
            }
            break;
          case OpCodes.HyperbolicSineIntegral: {
              double shi, chi;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.hyperbolicsinecosineintegrals(x, out shi, out chi);
                values[j] = shi;
              }
            }
            break;
          case OpCodes.HyperbolicCosineIntegral: {
              double shi, chi;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.hyperbolicsinecosineintegrals(x, out shi, out chi);
                values[j] = chi;
              }
            }
            break;
          case OpCodes.FresnelCosineIntegral: {
              double c = 0, s = 0;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.fresnelintegral(x, ref c, ref s);
                values[j] = c;
              }
            }
            break;
          case OpCodes.FresnelSineIntegral: {
              double c = 0, s = 0;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.fresnelintegral(x, ref c, ref s);
                values[j] = s;
              }
            }
            break;
          case OpCodes.AiryA: {
              double ai, aip, bi, bip;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.airy(x, out ai, out aip, out bi, out bip);
                values[j] = ai;
              }
            }
            break;
          case OpCodes.AiryB: {
              double ai, aip, bi, bip;
              var x = values[j - currentInstr.nArguments];
              if (double.IsNaN(x)) values[j] = double.NaN;
              else {
                alglib.airy(x, out ai, out aip, out bi, out bip);
                values[j] = bi;
              }
            }
            break;
          case OpCodes.Norm: {
              var x = values[j - currentInstr.nArguments];
              values[j] = double.IsNaN(x) ? double.NaN : alglib.normaldistribution(x);
            }
            break;
          case OpCodes.Erf: {
              var x = values[j - currentInstr.nArguments];
              values[j] = double.IsNaN(x) ? double.NaN : alglib.errorfunction(x);
            }
            break;
          case OpCodes.Bessel: {
              var x = values[j - currentInstr.nArguments];
              values[j] = double.IsNaN(x) ? double.NaN : alglib.besseli0(x);
            }
            break;
          case OpCodes.IfThenElse: {
              double condition = values[j - currentInstr.nArguments];
              double result = condition > 0.0 ? values[j - currentInstr.nArguments] : values[j - currentInstr.nArguments + 1];
              values[j] = result;
            }
            break;
          case OpCodes.AND: {
              double result = values[j - currentInstr.nArguments];
              for (int i = j - currentInstr.nArguments + 1; i < j; i++) {
                if (result > 0.0) result = values[i];
              }
              values[j] = result > 0.0 ? 1.0 : -1.0;
            }
            break;
          case OpCodes.OR: {
              double result = values[j - currentInstr.nArguments];
              for (int i = 1; i < currentInstr.nArguments; i++) {
                if (result <= 0.0) result = values[i]; ;
              }
              values[j] = result > 0.0 ? 1.0 : -1.0;
            }
            break;
          case OpCodes.NOT: {
              values[j] = values[j - currentInstr.nArguments] > 0.0 ? -1.0 : 1.0;
            }
            break;
          case OpCodes.GT: {
              double x = values[j - currentInstr.nArguments];
              double y = values[j - currentInstr.nArguments + 1];
              values[j] = x > y ? 1.0 : -1.0;
            }
            break;
          case OpCodes.LT: {
              double x = values[j - currentInstr.nArguments];
              double y = values[j - currentInstr.nArguments + 1];
              values[j] = x < y ? 1.0 : -1.0;
            }
            break;
          case OpCodes.TimeLag: {
              throw new NotSupportedException();
            }
          case OpCodes.Integral: {
              throw new NotSupportedException();
            }

          //mkommend: derivate calculation taken from: 
          //http://www.holoborodko.com/pavel/numerical-methods/numerical-derivative/smooth-low-noise-differentiators/
          //one sided smooth differentiatior, N = 4
          // y' = 1/8h (f_i + 2f_i-1, -2 f_i-3 - f_i-4)
          case OpCodes.Derivative: {
              throw new NotSupportedException();
            }
          case OpCodes.Call: {
              throw new NotSupportedException();
            }
          case OpCodes.Arg: {
              throw new NotSupportedException();
            }
          case OpCodes.Variable: {
              if (row < 0 || row >= dataset.Rows) values[j] = double.NaN;
              else {
                var variableTreeNode = (VariableTreeNode)currentInstr.dynamicNode;
                values[j] = ((IList<double>)currentInstr.iArg0)[row] * variableTreeNode.Weight;
              }
            }
            break;
          case OpCodes.LagVariable: {
              var laggedVariableTreeNode = (LaggedVariableTreeNode)currentInstr.dynamicNode;
              int actualRow = row + laggedVariableTreeNode.Lag;
              if (actualRow < 0 || actualRow >= dataset.Rows) values[j] = double.NaN;
              else {
                values[j] = ((IList<double>)currentInstr.iArg0)[actualRow] * laggedVariableTreeNode.Weight;
              }
            }
            break;
          case OpCodes.Constant: {
              var constTreeNode = (ConstantTreeNode)currentInstr.dynamicNode;
              values[j] = constTreeNode.Value;
            }
            break;

          //mkommend: this symbol uses the logistic function f(x) = 1 / (1 + e^(-alpha * x) ) 
          //to determine the relative amounts of the true and false branch see http://en.wikipedia.org/wiki/Logistic_function
          case OpCodes.VariableCondition: {
              if (row < 0 || row >= dataset.Rows) values[j] = double.NaN;
              else {
                var variableConditionTreeNode = (VariableConditionTreeNode)currentInstr.dynamicNode;
                double variableValue = ((IList<double>)currentInstr.iArg0)[row];
                double x = variableValue - variableConditionTreeNode.Threshold;
                double p = 1 / (1 + Math.Exp(-variableConditionTreeNode.Slope * x));

                double trueBranch = values[j - currentInstr.nArguments];
                double falseBranch = values[j - currentInstr.nArguments + 1];

                values[j] = trueBranch * p + falseBranch * (1 - p);
              }
            }
            break;
          default:
            throw new NotSupportedException();
        }
        #endregion

        // book-keeping in order to keep all the values correct
        int narg = currentInstr.nArguments;

        // we count the values before the next function is encountered
        if (narg == 0) { ++count; continue; }

        if (count > narg) {
          for (int i = 1; i <= count - narg; ++i)
            values[j - i] = values[j - i - narg];
        }

        count -= (narg - 1); // because after being evaluated the current instruction becomes a 'value'
      }

      return values[values.Length - 1];
    }
  }
}
